# -*- coding: utf-8 -*-
""" Прямая эмуляция алгоритмов sio2
Тест на ctypes находится в s4102@c4f1:tests/python/sio/sio.py
"""
import sys
pyver_ma = sys.version_info.major
pyver_mi = sys.version_info.minor
from collections import namedtuple
import os
import struct

def printf(s):
    print(s)
    return
# if pyver_ma == 3:
#     def printf(s):
#         print(s)
#         return
# else:
#     def printf(s):
#         print (s)
#         return

T_RECNAME = 1
T_GRPNAME = 2
T_POLNAME = 3
T_CATEXT  = 4
T_ARRDEF  = 5
T_ARRDIM  = 6

# typedef struct {
#   //  char catcpu_;		/* тип endian - не используется */
#   char  catver2;		/* '2' - версия для работы с большими данными */
#   short catmlen;		/* max длина каталога */
#   short catrlen;		/* занятая длина каталога */
#   short catdelkey_;		/* фикция  - использовать для endian */
#   long catfaddr;		/* фикция, в заголовок столько не записать */
#   short nexts;			/* число экстентов */
#   char date_y;			/* дата компиляции */
#   char date_m;			/* дата компиляции */
#   char date_d;			/* дата компиляции */
# } HCAT;

# struct cat {
#   char lcttype;			/* тип строки каталога: T_RECNAME ... 0-0 */
#   char lctname[23];		/* имя объекта (например записи)      1-24*/
#   long lctaddr;			/* адрес объекта в файле или номер поля при описании группы синоним lctnpol 25-32*/
#   long lctlen;			/* длина объекта в файле в байтах     33-40*/
#   int  lctform;			/* формат объекта XY (X - целое/вещ, Y - кол-во байт 41-44*/
#   int  lctnom;		        /* номер объекта (для экз. групп)     45-48*/
# };

st_HCAT = namedtuple('SIO_HCAT',['cat2ver',   # 1
                                 'catmlen',   # 2
                                 'catrlen',   # 2
                                 'catdelkey', # 2
                                 'catfaddr',  # 4
                                 'nexts',     # 2
                                 'date_y',    # 1
                                 'date_m',    # 1
                                 'date_d'])   # 1

st_MCAT = namedtuple('SIO_HCAT',['lcttype',   # 1
                                 'lctname',   # 23
                                 'lctaddr',   # 8
                                 'lctlen',    # 8
                                 'lctform',   # 4
                                 'lctnom'])   # 4
SIZEOFHCAT = 16
SIZEOFCAT  = 48

class SIO:
    """ Работа с файлами SIO2 на чтение
    """
    
    def __init__(self):
        self.f = None
        self.fname = None
        pass

    def siofop(self, fname):
        """ Открытие файла на чтение
        """
        rc = 0
        printf('Открывается файл "%s"' % fname)
        if self.fname:
            printf('Есть уже открытый файл "%s"' % self.fname)
            rc = -1
        else:
            self.fname = fname
            self.f = os.open(fname,0)
            self._read_hcat()
            #print(self.hcat)
            self._read_mcat2()
        return rc

    def siofcl(self):
        """
        """
        rc = 0
        printf('Закрывается файл "%s"' % self.fname)
        if self.f:
            os.close(self.f)
            self.fname = None
            self.h = None
        else:
            rc = -1
            printf('Файл еще не открыт')
        return rc

    def siofcat(self):
        """ Вывод каталога
        """
        printf ("----------------------------+------+----------------+------------")
        for mcatitem in self.mcat:
            lcttype = mcatitem.lcttype
            lctname = mcatitem.lctname
            lctaddr = mcatitem.lctaddr
            lctlen  = mcatitem.lctlen
            lctform = mcatitem.lctform
            lctnom  = mcatitem.lctnom
            if   lctform == 14: fmt, lfmt = 'real*4', 4
            elif lctform == 18: fmt, lfmt = 'real*8', 8
            elif lctform == 24: fmt, lfmt = 'int*4' , 4
            elif lctform == 28: fmt, lfmt = 'int*8' , 8
            else:               fmt, lfmt = '??',     1
            if   lcttype == T_RECNAME:
                if lctnom == 0:
                    printf("%-23s %3s %6s %10d %6s %12d" %
                          (lctname, '', 'record', lctlen/lfmt, fmt, lctaddr))
                else:
                    printf("%-23s %3d %6s %10d %6s %12d" %
                          (lctname, lctnom, 'egroup', lctlen/lfmt, 'bytes', lctaddr))
            elif lcttype == T_GRPNAME:
                printf("%-23s %3s %6s %10d %6s %5d fields" %
                      (lctname, '', 'group', lctlen, 'bytes', lctaddr))
            elif lcttype == T_POLNAME:
                printf("%-23s %3s %6s %10d %6s %12s" %
                      (lctname, '', 'field', lctlen/lfmt, fmt, ''))
            elif lcttype == T_CATEXT:
                printf("%-23s %3d %6s %10d %6s %12d" %
                       (lctname, lctnom, 'catal', lctlen, '', lctaddr))
            elif   lcttype == T_ARRDEF:
                printf("%-23s %2sD %6s %10d %6s %12d" %
                          (lctname, lctnom, 'array', lctlen/lfmt, fmt, lctaddr))
            elif   lcttype == T_ARRDIM:
                printf("%-23s %3d %6s %10d %6s %12d" %
                          (lctname, lctnom, 'array', lctlen/lfmt, fmt, lctaddr))
            else:
                printf ('@@ unknown type %d' % lcttype)
        printf("Длина файла:                                         %12d" % self.maxfaddr)
        printf("----------------------------+------+----------------+------------")
        return
    
    def siolencat(self):
        """ Выдача длины каталога
        """
        rc = len(self.mcat)
        return rc

    def siogetcat(self, ind):
        """ Выдача строки каталога по номеру
        """
        rc = self.mcat[ind]
        return rc
    
    def siorgetlen(self, name):
        """ Вывод ддины записи
        """
        rc = -1
        ind = self._FindInCat(name, T_RECNAME, 0)
        if ind < 0:
            return rc
        item = self.mcat[ind]
        if   item.lctform == 14: fmt, lfmt = 'real*4', 4
        elif item.lctform == 18: fmt, lfmt = 'real*8', 8
        elif item.lctform == 24: fmt, lfmt = 'int*4' , 4
        elif item.lctform == 28: fmt, lfmt = 'int*8' , 8
        else:                    fmt, lfmt = '??',     1
        rc = int(item.lctlen / lfmt)
        return rc

    def siorird(self, name, idx=0, lrec=0):
        """ Чтение записи со смещением заданной длины и указанным форматом
        """
        if not self.f:
            printf('siorird - файл еще не открыт')
            return []
        ind = self._FindInCat(name, T_RECNAME, 0)
        if ind < 0:
            printf ("#? siorird: Нет такого объекта: %s", name)
            return []
        #print('siorird', name, idx, lrec)
        item = self.mcat[ind]
        #print(item)
        frec = item.lctform
        if   item.lctform == 14: fmt, lfmt, fpack = 'real*4', 4, 'f'
        elif item.lctform == 18: fmt, lfmt, fpack = 'real*8', 8, 'd'
        elif item.lctform == 24: fmt, lfmt, fpack = 'int*4' , 4, 'i'
        elif item.lctform == 28: fmt, lfmt, fpack = 'int*8' , 8, 'q'
        else:                    fmt, lfmt, fpack = '??',     1, '?'
        if lrec == 0:
            lrdb = item.lctlen
        else:
            lrdb  = lrec * lfmt
        #print ('lrdb1=',lrdb, lrec)
        lidx = idx * lfmt
        print('siorird %s lctlen=%d, lidx=%d lrdb=%d' %
              (name, item.lctlen, lidx, lrdb))
        if lidx + lrdb > item.lctlen:
            lrdb = item.lctlen - lidx
        # print ('!! lrdn=%d' % lrdb)
        #print ('lrdb2=',lrdb)
        lrd = lrdb / lfmt
        if lrd < 0:
            printf("Превышен размер читаемых данных idx=%s lrd=%d" % (lidx, lrd))
            return []
        printf("read %s from %d len=%d bytes %d items" % (name, lidx, lrdb, lrd))
        os.lseek(self.f, item.lctaddr+lidx, os.SEEK_SET)
        b = os.read(self.f, lrdb)
        #print (rc)
        rc = []
        i = 0
        # while len(b) > 0:
        #     i += 1
        #     rc.append(struct.unpack(fpack,b[:lfmt])[0])
        #     b = b[lfmt:]
        #print('i=', i)
        return b
    
    def sioagetlen(self, name):
        """ Вывод ддины записи
        """
        rc = -1
        ind = self._FindInCat(name, T_ARRDEF, -1)
        if ind < 0:
            return rc
        item = self.mcat[ind]
        if   item.lctform == 14: fmt, lfmt = 'real*4', 4
        elif item.lctform == 18: fmt, lfmt = 'real*8', 8
        elif item.lctform == 24: fmt, lfmt = 'int*4' , 4
        elif item.lctform == 28: fmt, lfmt = 'int*8' , 8
        else:                    fmt, lfmt = '??',     1
        rc = int(item.lctlen / lfmt)
        return rc

    def sioagetdims(self, name):
        """ Чтение размерности массива
        """
        rc = 0
        if not self.f:
            printf('siorird - файл еще не открыт')
            return []
        ind = self._FindInCat(name, T_ARRDEF, -1)
        if ind < 0:
            print ("#? sioagetdims: Нет такого объекта: %s" % name)
            return 0
        item = self.mcat[ind]
        rc = item.lctnom
        return rc
    
    def sioagetdim(self, name, dim):
        """ Чтение длины измерения массива
        """
        rc = 0
        if not self.f:
            printf(u'siorird - файл еще не открыт')
            return []
        ind = self._FindInCat(name, T_ARRDIM, dim)
        if ind < 0:
            print (u"#? sioagetdim: Нет такого объекта: %s" % name)
            return 0
        item = self.mcat[ind]
        if   item.lctform == 14: fmt, lfmt, fpack = 'real*4', 4, 'f'
        elif item.lctform == 18: fmt, lfmt, fpack = 'real*8', 8, 'd'
        elif item.lctform == 24: fmt, lfmt, fpack = 'int*4' , 4, 'i'
        elif item.lctform == 28: fmt, lfmt, fpack = 'int*8' , 8, 'q'
        else:                    fmt, lfmt, fpack = '??',     1, '?'
        rc = item.lctlen / lfmt
        return rc
    
    def sioard(self, name, idx=0, lrec=0):
        """ Чтение массива ...
        """
        if not self.f:
            printf('sioard - файл еще не открыт')
            return []
        ind = self._FindInCat(name, T_ARRDEF, -1)
        if ind < 0:
            printf ("#? sioard: Нет такого объекта: %s", name)
            return []
        #print('siorird', name, idx, lrec)
        item = self.mcat[ind]
        #print('@@ sioard item=%s' % str(item))
        frec = item.lctform
        if   item.lctform == 14: fmt, lfmt, fpack = 'real*4', 4, 'f'
        elif item.lctform == 18: fmt, lfmt, fpack = 'real*8', 8, 'd'
        elif item.lctform == 24: fmt, lfmt, fpack = 'int*4' , 4, 'i'
        elif item.lctform == 28: fmt, lfmt, fpack = 'int*8' , 8, 'q'
        else:                    fmt, lfmt, fpack = '??',     1, '?'
        if lrec == 0:
            lrdb = item.lctlen
        else:
            lrdb  = lrec * lfmt
        #print ('lrdb1=',lrdb, lrec)
        lidx = idx * lfmt
        #print('@@ sioard %s lctlen=%d, lidx=%d lrdb=%d' %
        #      (name, item.lctlen, lidx, lrdb))
        if lidx + lrdb > item.lctlen:
            lrdb = item.lctlen - lidx
        # print ('!! lrdn=%d' % lrdb)
        #print ('lrdb2=',lrdb)
        lrd = lrdb / lfmt
        if lrd < 0:
            printf("Превышен размер читаемых данных idx=%s lrd=%d" % (lidx, lrd))
            return []
        #printf("@@ read %s from %d len=%d bytes %d items, faddr=%d" %
        #       (name, lidx, lrdb, lrd, item.lctaddr+lidx))
        os.lseek(self.f, item.lctaddr+lidx, os.SEEK_SET)
        b = os.read(self.f, lrdb)
        #print (rc)
        #rc = []
        #i = 0
        #while len(b) > 0:
        #    i += 1
        #    rc.append(struct.unpack(fpack,b[:lfmt])[0])
        #    b = b[lfmt:]
        #print('i=', i)
        return b
    
    def _read_hcat(self):
        """ Чтение заголовка
        """
        b = os.read(self.f,SIZEOFHCAT)
        #str = ''
        #for c in b: str += "%02x " % int(c)
        #print(str)
        t = struct.unpack('=bhhhihbbb',b)
        ver = chr(t[0])
        self.hcat = st_HCAT(ver,t[1],t[2],t[3],t[4],t[5],t[6],t[7],t[8])
        return

    def _read_mcat2(self):
        """ Чтение каталога
        """
        lcatt = self.hcat.catmlen * (self.hcat.nexts + 1)
        #print ('lcatt=',lcatt)
        self.maxfaddr = 16
        #print('---h maxfaddr=', self.maxfaddr, 16)
        self.mcat = []
        lencat = self.hcat.catrlen # столько надо прочитать
        for n in range(self.hcat.nexts+1):
            #print ('ext',n);
            if n > 0:           # ищем информацию об экстенте
                ind = self._FindInCat("## Extent ##",T_CATEXT,n);
                if ind < 0: 
                    printf("read_mcat ext=%d" % n); 
                    printf("#### Испорчен хвост каталога");
                    raise SystemExit
                #print ('ind=', ind)
                faddr = self.mcat[ind].lctaddr;
                os.lseek(self.f, faddr, os.SEEK_SET);
            #lwr = min(self.hcat.catmlen, lencat) * SIZEOFCAT
            lwr = self.hcat.catmlen * SIZEOFCAT
            self.maxfaddr += lwr;
            #print('---e maxfaddr=', self.maxfaddr, lwr)
            #print ('read',lwr)
            b = os.read(self.f, lwr)
            while len(b) > 0:
                if len(self.mcat) < self.hcat.catrlen:
                    t = struct.unpack('=b23bqqll',b[:SIZEOFCAT])
#                    lcttype = b[0]
                    lcttype = struct.unpack('=b',b[0:1])[0]
                    lctname = b[1:24].decode()
                    lctname = lctname.rstrip()
                    lctaddr = struct.unpack('=q',b[24:32])[0]
                    lctlen  = struct.unpack('=q',b[32:40])[0]
                    lctform = struct.unpack('=l',b[40:44])[0]
                    lctnom  = struct.unpack('=l',b[44:48])[0]
                    #print ('%d %-24s %d %12d %2d %12d' %
                    #       (lcttype, lctname, lctnom, lctlen, lctform, lctaddr))
                    mcatitem = st_MCAT(lcttype, lctname, lctaddr, lctlen,
                                       lctform, lctnom)
                    self.mcat.append(mcatitem)
                    if lcttype == T_RECNAME:
                        self.maxfaddr += lctlen
                else:           # пустая чать последнего экстента не переносится в каталог
                    #print('skip', len(self.mcat))
                    break
                b = b[SIZEOFCAT:]
            lencat -= self.hcat.catmlen
        return

    def _FindInCat(self, name, type, nom):
        ind = 0
        for item in self.mcat:
            #print ('!', name, type, nom, '-',
            #       item.lctname, item.lcttype, item.lctnom, ':',
            #       name == item.lctname)
            if item.lctname == name and item.lcttype == type and item.lctnom == nom:
                return ind
            if item.lctname == name and item.lcttype == type and -1 == nom:
                return ind
            ind += 1
        return -1

def PrintArray(a, ndims, sizes_str):
    sizes = []
    for s in sizes_str.split(): sizes.append(int(s))
#    print 'ndims = %d' % ndims
#    print 'sizes = %s' % sizes
    if   ndims == 1:
        sys.stdout.write("     00: ")
        firstI = True
        for i in range(sizes[0]):
            if i >= 2 and i < sizes[0]-2:
                if firstI: sys.stdout.write(".. ")
                firstI = False
                continue
            sys.stdout.write("%9.0f " % a[i])
        printf("")
    elif ndims == 2:
        firstI = True
        for i in range(sizes[0]):
            if i >= 1 and i < sizes[0]-1:
                #if firstI: sys.stdout.write("..i..\n")
                firstI = False
                continue
            sys.stdout.write("  %02d-00: "% i)
            firstJ = True
            for j in range(sizes[1]):
                if j >= 2 and j < sizes[1]-2:
                    if firstJ: sys.stdout.write(".. ")
                    firstJ = False
                    continue
                #print ("idx=%d len(a)=%d\n" % (i*sizes[1] + j, len(a)))
                sys.stdout.write("%9.0f " % a[i*sizes[1] + j])
            printf("")
    elif ndims == 3:
        firstI = True
        for i in range(sizes[0]):
            if i >= 1 and i < sizes[0]-1:
                #if firstI: sys.stdout.write("..i..\n")
                firstI = False
                continue
            firstJ = True
            for j in range(sizes[1]):
                if j >= 1 and j < sizes[1]-1:
                    #if firstJ: sys.stdout.write("..j..\n")
                    firstJ = False
                    continue
                #print ("idx=%d len(a)=%d\n" % (i*sizes[1] + j, len(a)))
                sys.stdout.write("%02d-%02d-00: "% (i, j))
                firstK = True
                for k in range(sizes[2]):
                    if k >= 2 and k < sizes[2]-2:
                        if firstK: sys.stdout.write(".. ")
                        firstK = False
                        continue
                    sys.stdout.write("%9.0f " % a[i*sizes[1]*sizes[2] + j*sizes[2] + k])
                printf("")
            #printf("j")
    else:
        print ('!! ndims %s = %d > 3' % (name, ndims))
        raise SystemExit
    return

if __name__ == '__main__':
    h = SIO()
    fname = '../tests/array.sio'
    #fname = '../tests/cnt_0001.001'
    #fname = '../KTPB/MultiPort-06-015-my.sio'
    #fname = '../KTPB/MultiPort-06-015.sio'
    #fname = '../KTPB/Out.sio'
    printf ('== siofop %s' % fname)
    h.siofop(fname)
    printf ('== siofcat')
    h.siofcat()
    #print ('== siocatlen')
    #lcat = h.siolencat()
    #print ('== siogetcat')
    #for i in range(lcat):
    #    item = h.siogetcat(i)
    #    print(item)
    for name in ['d1_1','d2_1','d3_1','d1_2','d2_2','d3_2',]:
        ndims = h.sioagetdims(name)
        sizes = ""
        alen = 1
        for dim in range(ndims):
            size = h.sioagetdim(name, dim)
            sizes += str(size) + ' '
            alen *= size
        print("== %s len=%7d ndims=%d sizes=%s" % (name, alen, ndims, sizes))
        res = h.sioard(name)
        PrintArray(res, ndims, sizes)
    printf ('== siofcl')
    h.siofcl()
    
