#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (C) vniitf, ignatiev
# просмотр записей и массивов в файле формата sio2
import sys
import struct
pyver_ma = sys.version_info.major
pyver_mi = sys.version_info.minor
#print ('##',pyver_ma, pyver_mi)
if pyver_ma == 3:
    import tkinter as tk
    from tkinter.dialog import Dialog
    from tkinter import filedialog
    from tkinter import commondialog
    from tkinter import ttk
    from tkinter.scrolledtext import ScrolledText
    import tkinter.font as tkfont
elif pyver_ma == 2:
    import Tkinter as tk
    import FileDialog as filedialog
    import ttk
    import tkFont as tkfont
else:
    print ('Unknown Python version %d.%d' % (pyver_ma, pyver_mi))
    raise SystemExit

import math

import sys
sys.path.append('/home/o37/sio2py/pysio')
from pysio import *

class ViewArrayN:
    """ Просмотр  массива в стиле HDFVIEW
    """
    
    NDIMS = ['I','J','K', 'L']  # это для понимания
    DIMS  = ['0','1','2', '3']  # так они в программе называются

    def __init__(self, root, h, name, fmt):
        self.root = root
        self.h = h
        self.name = name        # выбранный объект
        self.fmt   = fmt
        self.w = tk.Toplevel(root)
        self.w.title("%s %s" % (self.h.fname, self.name))
        self.w.protocol("WM_DELETE_WINDOW", self.onClose)

        self.cframe = tk.Frame(self.w)
        self.cframe.pack(fill='both', expand=True)
        self.b1frame = tk.Frame(self.w)
        self.b1frame.pack(fill=tk.X, expand=True)
        self.bframe = tk.Frame(self.b1frame)
        self.bframe.pack(side='left')
        self.tv = None          # treeview, бкдет создано позже в CreateTV
        self.sbx = self.sby = None # скроллеры
        self.popup = None          # для контекстного меню

        self.bframe.rowconfigure(0, weight=1)
        self.bframe.columnconfigure(0, weight=1)
        self.bframe.pack() #fill=tk.X, expand=True)
        #self.bframe.pack(fill=tk.X, expand=False)

        self.PrepairDims()
        if self.ndims == 1:
            self.w.geometry('400x600')
        else:
            self.w.geometry('1100x600')
        self.DoBFrame()
        self.tv = self.sbx = self.sby = None

        lFileName1 = tk.Label(self.bframe, text=u"Файл:")
        lFileName2= tk.Label(self.bframe, text=u"%s" % self.h.fname)
        lArrayName1 = tk.Label(self.bframe, text=u"Массив:")
        lArrayName2 = tk.Label(self.bframe, text=u"%s" % self.name)
        lFileName1.grid (row=0, column=0, sticky='w')
        lFileName2.grid (row=0, column=1, sticky='w')
        lArrayName1.grid(row=1, column=0, sticky='w')
        lArrayName2.grid(row=1, column=1, sticky='w')

        lfmt = tk.Label(self.bframe, text=self.fmt)
        lfmt.grid(row=1, column=2, sticky='w')
        lfull1 = tk.Label(self.bframe, text=str("Всего:"))
        lfull2 = tk.Label(self.bframe, text=str(str(self.rlen)))
        lfull1.grid(row=1, column=3, sticky='e')
        lfull2.grid(row=1, column=4, sticky='e')
        return
    
    def PrepairDims(self):
        """ Выбор информации о размерностях
        """
        print ('## PrepairDims')
        self.ndims = self.h.sioagetdims(self.name)
        self.cmap = ['']         # номера измерений для установки комбо
        for i in range(self.ndims):
            self.cmap.append(str(i))
        #print ('ndims=%d map=%s' % (self.ndims, self.cmap))
        self.mapped = []        # выбранные для просмотра измерения
        self.rlen = self.h.sioagetlen(self.name)
        self.mlen = [0]*self.ndims   # длины по измерениям (ilen, jlen, klen llen)
        self.idxb = [0]*self.ndims
        self.idxe = [0]*self.ndims
        for i in range(self.ndims):
            self.mlen[i] = self.h.sioagetdim(self.name,i)
            self.idxb[i] = 0
            self.idxe[i] = self.mlen[i]
        #print (self.mlen)
        #print (self.idxb)
        #print (self.idxe)
        return

    def DoBFrame(self):
        print ('## DoBFrame')
        """ Фрейм управления:
             0       |           1           |        2        |      3       |  4
     0  имя_файла    |                       |                 |              |
     1  Имя_массива  | формат                | число_элементов |              |
     2  комбо_1      | направление_1_и_длина | начало_вывода   | конец_вывода |
     3  комбо_2      | направление_2_и_длина | начало_вывода   | конец_вывода |
     4               | направление_3_и_длина | индекс          | <-1>         | <+1>
     5               | направление_4_и_длина | индекс          | <-1>         | <+1>
        """
        #bShow = tk.Button(self.bframe, text='Перевыдать', pady=1, command=self.Show)
        #bShow.grid(row=0, column=4, sticky='e')
        # строки с выбором маппинга
        nMapVars = 1
        if self.ndims > 1: nMapVars = 2
        self.vMap = []               # переменные для комбо выбора маппинга по 1 и 2 направлениям
        self.eMap = []               # комбобоксы для выбора маппинга по направлениям
        self.lidx = []               # информация о набравлении и его длине
        self.eb   = []               # поле ввода начального индекса вывода  по направлению
        self.ee   = []               # поле ввода конечного индекса вывода по направлению
        self.vb   = []               # переменные для начального индекса
        self.ve   = []               #  переменные для конечного индекса
        for n in range(nMapVars):
            dim = n             # напрвление для по умолчанию для маппинга
            self.vMap.append(tk.StringVar())
            self.eMap.append(ttk.Combobox(self.bframe, values=self.cmap, textvariable=self.vMap[n], width=2))
            self.eMap[n].bind('<<ComboboxSelected>>', lambda x,n=n: self.onMap(x,n)) # реакция на комюо
            self.eMap[n].current(1) # вывод умочательного значения
            self.eMap[n].grid(row=2+n, column=0)
            self.mapped.append(str(n))
            self.lidx.append(tk.Label(self.bframe, text='%s длина %d' % (self.NDIMS[n], self.mlen[n])))
            self.lidx[n].grid(row=2+n, column=1, sticky='w')
            self.vb.append(tk.StringVar()); self.vb[n].set(str(self.idxb[dim]))
            self.ve.append(tk.StringVar()); self.ve[n].set(str(self.idxe[dim]))
            #print('n=%d dim=%d %d->vb %d->ve' % (n, dim, self.idxb[dim], self.idxe[dim]))
            self.eb.append(tk.Entry(self.bframe,  textvariable=self.vb[n], width=8))
            self.ee.append(tk.Entry(self.bframe,  textvariable=self.ve[n], width=8))
            self.eb[n].grid(row=2+n, column=2)
            self.ee[n].grid(row=2+n, column=3)
            self.eb[n].bind('<KeyPress>', self.onEntryB)
            self.ee[n].bind('<KeyPress>', self.onEntryE)
            self.cmap.remove(str(n))
            #print('n=%d cmap=%s' % (n, self.cmap))
        
        if self.ndims < 3: return
        self.bl = [None]*self.ndims # кнопка <-1>
        self.br = [None]*self.ndims # кнопка <+1>
        #print('bl %d %s' % (len(self.bl), self.bl))
        for n in range(2, self.ndims): # теперь надо добавать следующие направления с кнопками +/-
            dim = n             # напрвление для по умолчанию для маппинга
            self.mapped.append(str(n))
            self.lidx.append(tk.Label(self.bframe, text='%s длина %d' % (self.NDIMS[n], self.mlen[n])))
            self.lidx[n].grid(row=2+n, column=1, sticky='w')
            self.vb.append(tk.StringVar()); self.vb[n].set(str(self.idxb[dim]))
            #print('n=%d dim=%d %d->vb %d->ve' % (n, dim, self.idxb[dim], self.idxe[dim]))
            self.eb.append(tk.Entry(self.bframe,  textvariable=self.vb[n], width=8))
            self.eb[n].grid(row=2+n, column=2)
            self.eb[n].bind('<KeyPress>', self.onEntryB)
            self.bl[n] = tk.Button(self.bframe, text='-1', pady=0, command=lambda n=n: self.onLeft(n))
            self.bl[n].grid(row=2+n, column=3)
            self.br[n] = tk.Button(self.bframe, text='+1', pady=0, command=lambda n=n: self.onRight(n))
            self.br[n].grid(row=2+n, column=4)
            self.cmap.remove(str(n))
            #print('n=%d cmap=%s' % (n, self.cmap))
        print('## DoBFrame final mapped=%s' % (self.mapped))
        self.Show()
        return

    def DoBounds(self):
        """ Формирование диапазонов выдачи
        """
        print('## DoBound')
        for ndim in range(self.ndims):
            mdim = self.__getDimByMap(ndim)
            #print('%d -> %d' % (ndim, mdim))
            if ndim < 2: # проверка начала и конца двух первых измерений
                vb = self.vb[ndim]
                idxb = int(vb.get())
                if idxb > self.mlen[mdim] - 10: idxb = self.mlen[mdim] - 10
                if idxb < 0:                    idxb= 0
                self.idxb[mdim] = idxb
                vb.set(str(idxb))
                ve = self.ve[ndim]
                #ve.set(str(idxe+99))
                idxe = int(ve.get())
                if (idxe - idxb) > 300:
                    idxe = idxb + 99 # неявно, по начальному. Что бы не было слишком длинных
                self.idxe[mdim] = idxe
                if idxe < idxb:             idxe = idxb + 99
                if idxe >= self.mlen[mdim]:
                    print('idxe1=%d mdim=%d mlem=%d'% (idxe, mdim, self.mlen[mdim]))
                    idxe = self.mlen[mdim]-1
                    print('idxe2=%d mdim=%d mlem=%d'% (idxe, mdim, self.mlen[mdim]))
                self.idxe[mdim] = idxe
                ve.set(str(idxe))
                #print('## DoBound ndim=%d mdim=%d idxb=%d idxe=%d' % (ndim, mdim, self.idxb[mdim], self.idxe[mdim]))
            else:               # проверка последующих изерений
                vb = self.vb[ndim]
                idxb = int(vb.get())
                if idxb > self.mlen[mdim] - 1: idxb = self.mlen[mdim] - 1
                if idxb < 0:                   idxb= 0
                self.idxb[mdim] = idxb
                vb.set(str(idxb))
        if self.ndims == 1:
            self.Ncol = 1
        else:
            mDim = self.__getDimByMap(1)
            self.Ncol = self.idxe[mDim] - self.idxb[mDim] + 1   # число колонок в таблице вывода
            #print(self.mlen)
            #print(self.idxb)
            #print(self.idxe)
        #print ('## DoBounds Ncol=%d' % self.Ncol)
        return

    def DoFormat(self, fmt):
        """ Преобразование формата из sio в текст и длину
        """
        lfmt = -1
        fpack = -1
        if   fmt == 'real*8': lfmt, fpack = (8,'d')
        elif fmt == 'real*4': lfmt, fpack = (4,'f')
        elif fmt == 'int*8' : lfmt, fpack = (8,'q')
        elif fmt == 'int*4' : lfmt, fpack = (4,'i')
        return (lfmt, fpack)

    def ConfigTV(self):
        """ Формирование второго стобца
        """
        print('## ConfigTV')
        colname = ''
        if self.ndims > 1:
            dim = self.__getDimByMap(1) # какое из направлений будет по строкам?
            for j in range(self.Ncol):
                colname += '%14d' % (j+self.idxb[dim]) + ' '
        else:
            colname += 'Значение'
        #print ('Ncol=%d colname=%s' % (self.Ncol, colname[:30]))
        (w,h) = (self.mfont.measure(' '*15), self.mfont.metrics("linespace"))
        self.witem = w
        (w,h) = (self.mfont.measure(' '*15*self.Ncol)+10, self.mfont.metrics("linespace"))
        #print ('Ncol=%d w=%d' % (self.Ncol, w))
        self.wrow = w
        self.tv.column("c2", minwidth = w)
        self.tv.heading('c2', text = colname, anchor = tk.W)
        return
    
    def CreateTV(self):
        """ Пересоздание treeview
        """
        if self.tv:  self.tv.destroy()
        if self.sbx: self.sbx.destroy()
        if self.sby: self.sby.destroy()
        #print ('===== Создаем TV и SB')
        self.tv = ttk.Treeview(self.cframe, height=self.Nrow)
        style = ttk.Style()
        font = ('monospace',8)
        self.mfont = tkfont.Font(family=font[0], size=font[1])
        style.configure('Treeview.Heading', font=font)
        style.configure('Treeview', font=font)

        self.tv.bind("<ButtonPress-1>", self.onMouseP)
        self.tv.bind("<ButtonRelease-1>", self.onMouseR)

        self.tv.grid(row = 0, column = 0, sticky='nsew')
        self.tv['columns']=('c1','c2')
        self.tv.column('#0', width=0, stretch=tk.NO)
        self.tv.column("c1", minwidth = 80)
        self.tv.heading('#0', text='', anchor=tk.CENTER)
        self.tv.heading('c1', text = self.c1text, anchor = tk.CENTER)
        self.ConfigTV()
        
        # Scrollbars + attach scrollbars to TreeView
        sby = tk.Scrollbar(self.cframe, orient = "vertical", command = self.tv.yview)
        sbx = tk.Scrollbar(self.cframe, orient = "horizontal", command = self.tv.xview)
        self.tv.configure(yscrollcommand = sby.set, xscrollcommand = sbx.set)
        sby.grid(row = 0, column = 1, sticky = "ns")
        sbx.grid(row = 1, column = 0, sticky = "ew")

        # Configure and position grid for TreeView
        self.cframe.grid_rowconfigure(0, weight = 1)
        self.cframe.grid_columnconfigure(0, weight = 1)
        return

    def onEntryB(self, event):
        """ Реакция на ввод в поле ввода начала
        """
        if event.keysym == "Return":
            self.Show()
        return

    def onEntryE(self, event):
        """ Реакция на ввод в поле ввода конца
        """
        if event.keysym == "Return":
            self.Show()
        return

    def onMap(self, e, ndim):
        """ Реакция на работу с комбобоксом
        ndim - номер строки (графическое измерение в програме)
        mdim - выбранное новое реальное измерение массива для выборки границ и формирования индекса
        """
        n = self.vMap[ndim].get()
        if not n:               # повторный маппинг
            self.__reset()
            return
        mdim = int(n)            # номер выбранного направления длы маппинга mdim -> ndim
        print ("## onMap Выбрано для %s: %s" % (self.NDIMS[ndim], mdim))
        if ndim == 0: # меняем первое измерение, при этом надо будет перевыбрать второе и переустановить следующие
            self.cmap = [''] #, 'I','J','K','L']
            for i in range(self.ndims):
                if str(i) != str(mdim): self.cmap.append(str(i))
            self.mapped = [str(mdim)]
            self.eMap[1]['values'] = self.cmap
            self.eMap[1].current(0)
            l, idxb, idxe = self.__MakeMapBound(mdim)
            self.lidx[0]['text'] = '%s длина %d' % (self.NDIMS[mdim], l)
            self.vb[0].set(str(idxb))
            for i in range(1, self.ndims):
                self.lidx[i]['text'] = ''
                self.vb[i].set('')
            print ('-- ndim=%s mdim=%s %s' % (ndim, mdim, self.NDIMS[mdim]))
            print('ndim=%d mapped=%s cmap=%s' % (ndim, self.mapped, self.cmap))
        else:                   # ndim = 1, переделываем спикок возможных направлений
            if len(self.mapped) == 1: self.mapped.append(str(mdim))
            else:                     self.mapped[1] = str(mdim)
            l, idxb, idxe = self.__MakeMapBound(mdim)
            self.lidx[1]['text'] = '%s длина %d' % (self.NDIMS[mdim], l)
            self.vb[1].set(str(idxb))
            print ('-- ndim=%s mdim=%s %s' % (ndim, mdim, self.NDIMS[mdim]))
            print('ndim=%d mapped=%s cmap=%s' % (ndim, self.mapped, self.cmap))
            if self.ndims > 2: # теперь надо пределать все последующие направления
                self.mapped = self.mapped[:2]
                for i in range(self.ndims):
                    if str(i) != self.mapped[0] and str(i) != self.mapped[1]:
                        self.mapped.append(str(i))
                for ndim in range(2,self.ndims):
                    mdim = self.__getDimByMap(ndim) # номер выбранного направления длы маппинга mdim -> ndims
                    l, idxb, idxe = self.__MakeMapBound(mdim)
                    self.lidx[ndim]['text'] = '%s длина %d' % (self.NDIMS[mdim], l)
                    self.vb[ndim].set(str(idxb))
                    print ('++ ndim=%s mdim=%s %s' % (ndim, mdim, self.NDIMS[mdim]))

        print('## onMap final mapped=%s' % (self.mapped))
        if len(self.mapped) == self.ndims: self.Show() # маппинг сделан, можно перевыдовать
        return
        
    def onLeft(self, ndim):
        """ Уменьшение по ndim
        """
        mdim = self.__getDimByMap(ndim)
        print ("## onLeft %d -> %d" % (ndim, mdim))
        self.idxb[mdim] -= 1
        if self.idxb[mdim] < 0: self.idxb[mdim] = 0
        #print ("onLeft mdim=%d %d" % (mdim, self.idxb[mdim]))
        self.vb[ndim].set(str(self.idxb[mdim]))
        print(self.idxb)
        self.Show()
        return

    def onRight(self, ndim):
        """ Увеличение по ndim
        """
        mdim = self.__getDimByMap(ndim)
        print ("## onRight %d - >%s" % (ndim, mdim))
        mlen = self.mlen[mdim]
        self.idxb[mdim] += 1
        if self.idxb[mdim] > mlen - 1: self.idxb[mdim] = mlen - 1
        if self.idxb[mdim] < 0: self.idxb[mdim] = 0
        #print ("onRight mdim=%d %d mlen=%d" % (mdim, self.idxb[mdim], mlen))
        self.vb[ndim].set(str(self.idxb[mdim]))
        print(self.idxb)
        self.Show()
        return

    def onMouseP(self, event):
        """ Реакция на мышь 1
        """
        iid = self.tv.identify_row(event.y)
        #printf ('onMouseP %s iid=%s' % (event, iid))
        if iid:
            self.tv.selection_set(iid)
            idx = self.tv.selection()[0]
            #print('idx=%s' % idx)
            item = self.tv.item(idx)
            if item:
                r = item["values"]
                x, y = event.x_root, event.y_root
                self.popup = tk.Toplevel(None)
                self.popup.wm_overrideredirect(True)
                self.popup.wm_geometry("+%d+%d" % (x, y))
                rr = r[0].split(':')
                #print('rr=%s x=%d y=%d witem=%d' %
                #      (rr, x, y, self.witem))
                idxs = ''
                for jj in range(self.ndims):
                    mdim = self.__getDimByMap(jj)
                    dim = self.NDIMS[mdim]
                    idx = rr[jj]
                    if jj == 1: idx = '*' # чтобы просчитать этот индекс, надо учитывать скроллинг
                    idxs += '%s:%s ' % (dim, idx)
                idxs = idxs[:-1]
                text = u'Индексы: %s' % (idxs)
                bg = "#dddd40"
                label = tk.Label(self.popup, text=text, justify='left',
                                 background=bg, relief='solid', borderwidth=1)
                #wraplength = self.wraplength)
                label.pack(ipadx=1)
        return

    def onMouseR(self, event):
        """ Реакция на мышь 1
        """
        iid = self.tv.identify_row(event.y)
        #printf ('onMouseR %s iid=%s' % (event, iid))
        if self.popup: self.popup.destroy()
        self.popup = None
        return

    def Show(self):
        print '## Show' % (self.mapped)
        self.DoBounds()
        if len(self.mapped) < self.ndims: return
        lfmt, fpack = self.DoFormat(self.fmt)
        if   lfmt < 0:
            print ('Uknown format %s for %s' % (self.fmt, self.name))
            return
        printf ("ViewArray-%d.Show name=%s file=%s" % (
            self.ndims,
            self.name,
            self.h.fname,
        ))
        self.c1text = ''
        for i in range(self.ndims):
            self.c1text += '%s:' % self.NDIMS[int(self.mapped[i])]
        self.c1text = self.c1text[:-1]
        #print 'c1text=%s' % self.c1text
        mlen = [0] * self.ndims
        idxe = [0] * self.ndims
        idxb = [0] * self.ndims
        for i in range(len(self.mapped)):
            mdim = self.__getDimByMap(i)
            mlen[i] = self.mlen[mdim]
            idxb[i] = self.idxb[mdim]
            idxe[i] = self.idxe[mdim]
        #print('mlen=%s - %s' % (mlen, self.mlen))
        #print('idxb=%s - %s' % (idxb, self.idxb))
        #print('idxe=%s - %s' % (idxe, self.idxe))
        self.Nrow = mlen[0]   # число строк в таблице вывода
        if self.Nrow > 20: self.Nrow = 20
        self.CreateTV()
        b = self.h.sioard(self.name, 0, 0)
        nrow = idxe[0] - idxb[0] + 1
        #print('nrow=%d' % nrow)
        adim = [None]*self.ndims # адрес для высвечивания с учетом маппинга
        m2n  = [None]*self.ndims # индекс для i,j,k... в mappped
        #n2m  = [None]*self.ndims # номера mdim в mapped по порядку
        for jj in range(self.ndims):
            m2n[jj] = self.mapped.index(str(jj))
            #n2m[self.mapped.index(str(jj))] = jj
        #print('MAP: m2n:%s n2m:%s' % (m2n, n2m))
        
        for row in range(nrow): # row = I, строка по J
            adim[0] = row + idxb[0] # начала вырезок по измерениям
            for jj in range(1, self.ndims):
                adim[jj] = idxb[jj]
            addr = ""           # адрес начала строки в массиве
            for jj in range(self.ndims):
                addr += '%d:' % (adim[jj])
            addr = addr[:-1]
            c = ''
            for col in range(self.Ncol):
                #idx = ((i*self.mlen[1] + j)*self.mlen[2] + k)* self.mlen[3] + l
                idx = adim[m2n[0]]
                for mDim in range(1,self.ndims):
                    idx = idx * mlen[m2n[mDim]] + adim[m2n[mDim]]
                #if row < 2 and col < 2:
                #    print("row=%d col=%d adim=%s idx=%d" % (row, col, adim, idx))
                if idx >= len(b)/lfmt : c += '+'*14 + ' '
                else:
                    bvalue = b[idx*lfmt:(idx+1)*lfmt]
                    value  = struct.unpack(fpack,bvalue)[0]
                    c += "%14.8g" % value + ' '
                if self.ndims > 1: adim[1] += 1
            self.tv.insert(parent='', index=row, iid=row, text='',
                           values=(addr, c))
        return
    
    def onClose(self):
        """ Закрытие окна
        """
        printf ("ViewArrayN.onClose")
        self.w.destroy()
        return
    
    def __MakeMapBound(self, n):
        """ Формирование границ по измерениею при маппинге
        """
        idx = int(n)
        l    = self.mlen[idx]
        idxb = self.idxb[idx]
        idxe = self.idxe[idx]
        return l, idxb, idxe

    def __reset(self):
        """ Сброс и повторное формирование маппинга
        """
        self.cmap = [''] #, 'I','J','K','L']
        for i in range(self.ndims):
            self.cmap.append(str(i))
        self.mapped = []
        lm = self.ndims
        if lm > 2: lm = 2       # только два направления маппируются
        for i in range(lm):
            self.eMap[i]['values'] = self.cmap
            self.eMap[i].current(0)
        print('Reset', self.cmap, self.mapped)
        return 

    def __getDimByMap(self, dim):
        """ Перевод графического измерения в реальное
        """
        #print("## __getDimByMap mdim=%d mapped=%s" % (dim, self.mapped))
        if len(self.mapped) <= dim:
            print '?'*100
            return
        m = self.mapped[dim]
        mdim = int(m)
        return mdim
    
class SIOVIEWRECORD(tk.Frame):
    """ Просмотр записи
    """
    
    def __init__(self, root, h, name, fmt):
        self.root = root
        self.h = h
        self.name = name        # выбранный объект
        self.fmt  = fmt
        # print(self.name)
        self.w = tk.Toplevel(root)
        self.w.title(self.name)
        self.w.protocol("WM_DELETE_WINDOW", self.onClose)
        self.cframe = tk.Frame(self.w)
        self.bframe = tk.Frame(self.w)
        self.cframe.pack()
        self.bframe.pack()
        
        self.tv = ttk.Treeview(self.cframe, height=20)
        self.tv['columns']=('addr', 'c0', 'c1', 'c2', 'c3', 'c4','c5','c6','c7','c8','c9')
        w1 = 100
        w2 = 100
        self.tv.column('#0' , width=0, stretch=tk.NO)
        self.tv.column('addr', width=w1, anchor=tk.W)
        for c in ['c0','c1','c2','c3','c4','c5','c6','c7','c8','c9']:
            self.tv.column(c, width=w2, anchor=tk.W)

        self.tv.heading('#0' , text=''         , anchor=tk.CENTER)
        self.tv.heading('addr', text='Адрес'     , anchor=tk.CENTER)
        for c in ['c0','c1','c2','c3','c4','c5','c6','c7','c8','c9']:
            self.tv.heading(c, text=c[1:], anchor=tk.CENTER)
        #self.tv.bind("<Button-3>", self.onMouse)
        #self.tv.bind("<Double-Button-1>", self.onD1)
        self.tv.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        
        self.sb = tk.Scrollbar(self.cframe)
        self.sb.pack(side=tk.RIGHT, fill=tk.Y, expand=False)

        self.tv.config(yscrollcommand=self.sb.set)
        self.sb.config(command=self.tv.yview)
        self.idx = 0
        self.len = 2000
        self.rlen = self.h.siorgetlen(self.name)
        self.ve1 = tk.StringVar(); self.ve1.set(str(self.idx))
        self.ve2 = tk.StringVar(); self.ve2.set(str(self.len))
        self.ve1.set(str(self.idx))
        self.ve2.set(str(self.len))
        lidx = tk.Label(self.bframe, text='от')
        self.e1 = tk.Entry(self.bframe,  textvariable=self.ve1)
        llen = tk.Label(self.bframe, text='длиной')
        self.e2 = tk.Entry(self.bframe,  textvariable=self.ve2)
        lfull = tk.Label(self.bframe, text=str("Всего " + str(self.rlen)))
        lidx.pack(side='left')
        self.e1.pack(side='left')
        llen.pack(side='left')
        self.e2.pack(side='left')
        lfull.pack(side='left')
        bShow = tk.Button(self.bframe, text='Перевыдать', command=self.ShowRecord)
        bShow.pack(side='right')
        return

    def ShowRecord(self):
        """ Вывод записи
        """
        lfmt = 0
        fpack = 0
        if   self.fmt == 'real*8': lfmt, fpack = (8,'d')
        elif self.fmt == 'real*4': lfmt, fpack = (4,'f')
        elif self.fmt == 'int*8' : lfmt, fpack = (8,'q')
        elif self.fmt == 'int*4' : lfmt, fpack = (4,'i')
        else:
            print ('Uknown format %s for %s' % (self.fmt, self.name))
            return
        self.idx = int(self.ve1.get())
        self.len = int(self.ve2.get())
        printf ("SIOVIEWRECORD.ShowRecord name=%s file=%s %d:%d" % (
            self.name,
            self.h.fname,
            self.idx, self.len))
        b = self.h.siorird(self.name, self.idx, self.len)
        #print ('len b=', len(b)/lfmt)
        #if b:
        #    for i in b: print(i)
        for i in self.tv.get_children():
            self.tv.delete(i)
        nrow = int((len(b)/lfmt - 1) / 10) + 1
        #n = int(math.log(self.rlen,10)+0.000000001) +1 # число цифр в адресе
        n = int(round(math.log(self.rlen,10)) + 1)
        for row in range(nrow):
            a1 = a2 = self.idx + (row * 10)
            i = self.idx + row * 10
            #print('row=', row, 'idx=', self.idx, 'i=',i, self.len, n)
            c = ['']*10
            for j in range(10):
                #printf('row=%d i=%d j=%d lb=%d %s' % (row, i, j, len(b), row * 10 + j < len(b)))
                if row * 10 + j < len(b)/lfmt:
                    value = b[(row * 10 + j)*lfmt:(row * 10 + j+1)*lfmt]
                    c[j] = struct.unpack(fpack,value)[0]
#                    c[j] = b[row * 10 + j]; a2 = a1 + j
                else:
                    c[j] = ''
            #print a1, type(a1), a2, type(a2), n
            addr = '%*d:%*d' % (n, a1, n, a2)
            self.tv.insert(parent='', index=row, iid=row, text='',
                           values=(addr, c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], c[8], c[9]))
        return
    
    def onClose(self):
        """ Закрытие окна
        """
        printf ("SIOVIEWRECORD.onClose")
        self.w.destroy()
        return
    
class SIOVIEWFILE(tk.Frame):
    """ Просмотр каталога файла
    """

    def __init__(self, root, datafile):
        self.root = root
        self.w = tk.Toplevel(root)
        self.w.title(datafile)
        self.w.protocol("WM_DELETE_WINDOW", self.onClose)
        self.cframe = tk.Frame(self.w)
        self.cframe.pack(fill='both', expand=True)
        
        self.h = SIO()
        self.tv = ttk.Treeview(self.cframe, height=20)
        self.tv['columns']=('c1', 'c2', 'c3', 'c4','c5','c6','c7')
        w1 = 200
        w2 = 30
        w3 = 60
        w4 = 100
        w5 = 60
        w6 = 100
        w7 = 100
        self.tv.column('#0' , width=0, stretch=tk.NO)
        self.tv.column('c1', width=w1, anchor=tk.W)
        self.tv.column('c2', width=w2, anchor=tk.W)
        self.tv.column('c3', width=w3, anchor=tk.W)
        self.tv.column('c4', width=w4, anchor=tk.E)
        self.tv.column('c5', width=w5, anchor=tk.W)
        self.tv.column('c6', width=w6, anchor=tk.E)
        self.tv.column('c7', width=w7, anchor=tk.E)

        self.tv.heading('#0' , text='',            anchor=tk.CENTER)
        self.tv.heading('c1', text='Имя',          anchor=tk.CENTER)
        self.tv.heading('c2', text='#',            anchor=tk.CENTER)
        self.tv.heading('c3', text='Тип',          anchor=tk.CENTER)
        self.tv.heading('c4', text='Длина',        anchor=tk.CENTER)
        self.tv.heading('c5', text='Формат',       anchor=tk.CENTER)
        self.tv.heading('c6', text='Адрес',        anchor=tk.CENTER)
        self.tv.heading('c7', text='Размерность' , anchor=tk.CENTER)
        self.tv.bind("<Button-3>", self.onMouse)
        self.tv.bind("<Double-Button-1>", self.onD1)
        self.tv.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        
        self.sb = tk.Scrollbar(self.cframe)
        self.sb.pack(side=tk.RIGHT, fill=tk.Y, expand=False)
        self.tv.config(yscrollcommand=self.sb.set)
        self.sb.config(command=self.tv.yview)

        self.h.siofop(datafile)
        self.items = []         # объекты файла
        # self.h.siofcat()
        self.cat()
        self.menur = tk.Menu(tearoff=0)
        self.menur.add_command(label='Показать' , command=self.onShowRecord)
        self.menua = tk.Menu(tearoff=0)
        self.menua.add_command(label='Показать' , command=self.onShowArray)
        return
    
    def cat(self):
        """ Вывод каталога файла
        """
        for i in self.tv.get_children():
            self.tv.delete(i)
        nitems = self.h.siolencat()
        arraydef = None         # для последующего заполнения размерностей
        for idx in range(nitems):
            item = self.h.siogetcat(idx)
            self.items.append(item)
            #print (item)
            c1 = item.lctname
            c7 = ''
            if item.lctnom == 0: c2 = ''
            else:                c2 = str(item.lctnom)
            if   item.lcttype == 1 and item.lctnom == 0: c3 = 'Record'
            elif item.lcttype == 1 and item.lctnom != 0: c3 = 'eGroup'
            elif item.lcttype == 2: c3 = 'Group'
            elif item.lcttype == 3: c3 = 'Field'
            elif item.lcttype == 4: c3 = 'Extend'
            elif item.lcttype == 5: c3 = 'Array'
            elif item.lcttype == 6: c3 = 'Dim'
            else:                   c3 = item.lcttype
            if   item.lctform == 14: c5, lfmt = 'real*4', 4
            elif item.lctform == 18: c5, lfmt = 'real*8', 8
            elif item.lctform == 24: c5, lfmt = 'int*4',  4
            elif item.lctform == 28: c5, lfmt = 'int*8',  8
            else:                    c5, lfmt = item.lctform, 1
            c4 = item.lctlen / lfmt
            c6 = item.lctaddr
            if c3 == 'Array':
                c2 += 'D'
                c7 = ''
                arraydef = [c1,c2,c3,c4,c5,c6,c7]
            if c3 == 'Dim'  :
                c2 = str(item.lctnom)
                dim = item.lctnom
                size = str(item.lctlen / lfmt)
                ndims = int(arraydef[1][:-1]) # без D
                sizes = arraydef[6]
                if dim == ndims-1:
                    sizes += size
                else:
                    sizes += size + 'x'
                arraydef[6] = sizes
                #print ("%s dim=%d of %s" % (arraydef, dim, ndims))
                if dim == ndims-1:
                    c1, c2, c3, c4, c5, c6, c7 = arraydef
                    self.tv.insert(parent='', index=idx, iid=idx, text='',
                               values=(c1, c2, c3, c4, c5, c6, c7))
            if c3 != 'Extend'  and c3 !='Dim' and c3 != 'Array':
                self.tv.insert(parent='', index=idx, iid=idx, text='',
                               values=(c1, c2, c3, c4, c5, c6, c7))
        return

    def onShowArray(self):
        """ Вывод массива ala hdfview
        """
        name, fmt = self.selected
        ndims = self.h.sioagetdims(name)
        printf ('## onShowArray %s ndims=%d fmt=%s' % (name, ndims, fmt))

        varray = ViewArrayN(self.root, self.h, name, fmt)
        varray.Show()
        return

    def onShowRecord(self):
        """ Вывод записи
        """
        name, fmt = self.selected
        printf ('## onShowRecord %s' % name)
        vrecorrd = SIOVIEWRECORD(self.root, self.h, name, fmt)
        vrecorrd.ShowRecord()
        return
    
    def onClose(self):
        """ Закрытие файла и окна
        """
        printf('## onClose')
        if self.h and self.h.f:
            self.h.siofcl()
        self.w.destroy()
        return

    def contextMenuR(self, x,y):
        """ Контекстное меню
        """
        self.menuR.post(x,y)
        return

    def contextMenuA(self, x,y):
        """ Контекстное меню
        """
        self.menua.post(x,y)
        return

    def onMouse(self, event):
        """ Реакция на мышь 3
        """
        iid = self.tv.identify_row(event.y)
        printf ('## onMouse %s iid=%s' % (event, iid))
        if iid:
            self.tv.selection_set(iid)
            idx = self.tv.selection()[0]
            #print('idx=%s' % idx)
            item = self.tv.item(idx)
            r = item["values"]
            #printf('item=%s' % (item))
            #printf('r=%s' % (r))
            t,name,nom,fmt = (r[2],r[0],r[1],r[4])
            if t == 'Record' and nom == '':
                self.selected = name, fmt
                self.contextMenuR(event.x_root, event.y_root)
            if t == 'Array':
                self.selected = name, fmt
                self.contextMenuA(event.x_root, event.y_root)
            else:
                print ('Данный тип объекта пока не выдается: %s %s %s' % (
                    str(t), str(name), str(nom)))
        return
    
    def onD1(self, event):
        """ Реакция на двойную мышь 1
        """
        iid = self.tv.identify_row(event.y)
        printf ('## onD1 %s iid=%s' % (event, iid))
        if iid:
            self.tv.selection_set(iid)
            idx = self.tv.selection()[0]
            #print('idx=%s' % idx)
            item = self.tv.item(idx)
            #print('item=', item)
            r = item["values"]
            #printf('item=%s' % (item))
            #printf('r=%s' % (r))
            t,name,nom, fmt = (r[2],r[0],r[1],r[4])
            if t == 'Record' and nom == '':
                self.selected = (name, fmt)
                self.onShowRecord()
            elif t == 'Array':
                self.selected = name, fmt
                self.onShowArray()
            else:
                print ('Данный тип объекта пока не выдается: %s %s %s' % (
                    str(t), str(name), str(nom)))
        return

class SIOVIEW:

    def __init__(self, root):
        self.root = root
        cframe = tk.Frame()
        bframe = tk.Frame()
        cframe.pack(side='top', expand=True, fill=tk.BOTH)
        label = tk.Label(cframe, text="""Просмотр данных в файле формата sio2
(C) ВНИИТФ, Игнатьев
""")
        label.pack()
        bframe.pack(side='top')
        bQuit = tk.Button(bframe, text='Выход', command=self.onQuit)
        bOpen = tk.Button(bframe, text='Открыть', command=self.onOpen)
        bQuit.pack(side='left')
        bOpen.pack(side='right')
        return
        
    def onOpen(self):
        print('## onOpen')
        fd = filedialog.LoadFileDialog(root)
        datafile = fd.go(key="/")
        print ('f=', datafile)
        if datafile:
            sio = SIOVIEWFILE(root, datafile)
        return

    def onQuit(self):
        printf('## onQuit')
        self.root.destroy()
        return

if __name__ == '__main__':
    root = tk.Tk()
    root.geometry('300x80+0+0')
    root.title('sioview')
    view = SIOVIEW(root)
    if len(sys.argv) > 1:
        print (sys.argv)
        datafile = sys.argv[1]
        sio = SIOVIEWFILE(root, datafile)
    root.mainloop()
